package de.tudresden.inf.st.mquat.solving.clustering;

import de.tudresden.inf.st.mquat.jastadd.model.Resource;
import de.tudresden.inf.st.mquat.jastadd.model.Root;
import weka.clusterers.*;
import weka.core.Attribute;
import weka.core.Instances;

import java.util.*;

/***
 * Wrapper for WEKA SimpleKMeans. Clustering based on:
 * CPU frequency
 * RAM total
 * Disk total
 * Network throughput
 *
 * @author Dmytro Pukhkaiev
 * @version 0.1
 */
public class KMeans{
    // dataset
    private Instances resources;
    // number of clusters
    private int k;
    // list of resources
    private ArrayList<HardwareComponent> hardwareComponentList = new ArrayList<>();

    /***
     * Prepare clustering algorithm, extract WEKA instances from Jastadd model
     * @param model Jastadd model
     */
    public KMeans(Root model) {
        // create attributes for instances
        Attribute cpu = new Attribute("cpu");
        Attribute ram = new Attribute("ram");
        Attribute disk = new Attribute("disk");
        Attribute network = new Attribute("network");
        ArrayList<Attribute> attributeList = new ArrayList<Attribute>();
        attributeList.add(cpu);
        attributeList.add(ram);
        attributeList.add(disk);
        attributeList.add(network);

        // create a dataset
        resources = new Instances("resources", attributeList, 0);

        //instantiate items in the dataset
        for (Resource resource : model.getHardwareModel().getResourceList()) {
            if (resource.getType().getName().getName().equals("ComputeNode")) {
                HardwareComponent hardwareComponent = new HardwareComponent(resource, resource.getNumSubResource());
                hardwareComponent.getInstance().setDataset(resources);

                for (Resource sub : resource.getSubResourceList()) {
                    String typeName = sub.getType().getName().getName();
                    switch (typeName) {
                        case "CPU":
                            hardwareComponent.getValue().put("CPU", sub.getCurrentValueByPropertyName("frequency"));
                            //hardwareComponent.getInstance().setValue(cpu, sub.getCurrentValueByPropertyName("frequency"));
                            break;
                        case "RAM":
                            hardwareComponent.getValue().put("RAM", sub.getCurrentValueByPropertyName("total"));
                            //hardwareComponent.getInstance().setValue(ram, sub.getCurrentValueByPropertyName("total"));
                            break;
                        case "DISK":
                            hardwareComponent.getValue().put("DISK", sub.getCurrentValueByPropertyName("total"));
                            //hardwareComponent.getInstance().setValue(disk, sub.getCurrentValueByPropertyName("total"));
                            break;
                        case "NETWORK":
                            hardwareComponent.getValue().put("NETWORK", sub.getCurrentValueByPropertyName("throughput"));
                            //hardwareComponent.getInstance().setValue(network, sub.getCurrentValueByPropertyName("throughput"));
                            break;
                        default:
                            throw new RuntimeException("Unable to transform unknown resource of type" + typeName);
                    }
                }

                hardwareComponentList.add(hardwareComponent);
            }
        }

        Map<String, Double> minValue = new HashMap<>();
        Map<String, Double> maxValue = new HashMap<>();

        for (HardwareComponent hwc :hardwareComponentList){
            for (String key : hwc.getValue().keySet()){
                if (! minValue.containsKey(key))
                    minValue.put(key, hwc.getValue().get(key));
                else if (minValue.get(key) > hwc.getValue().get(key))
                    minValue.put(key, hwc.getValue().get(key));

                if (! maxValue.containsKey(key))
                    maxValue.put(key, hwc.getValue().get(key));
                else if (maxValue.get(key) < hwc.getValue().get(key))
                    maxValue.put(key, hwc.getValue().get(key));
            }
        }
        LinearNormalizer linearNormalizer = new LinearNormalizer(minValue, maxValue);

        for (HardwareComponent hwc : hardwareComponentList) {
            hwc.setNormalizedValue(linearNormalizer.normalize(hwc));
            for (String key : hwc.getNormalizedValue().keySet()){
                switch (key){
                    case "CPU":
                        hwc.getInstance().setValue(cpu, hwc.getNormalizedValue().get(key));

                        break;
                    case "RAM":
                        hwc.getInstance().setValue(ram, hwc.getNormalizedValue().get(key));
                        break;
                    case "DISK":
                        hwc.getInstance().setValue(disk, hwc.getNormalizedValue().get(key));
                        break;
                    case "NETWORK":
                        hwc.getInstance().setValue(network, hwc.getNormalizedValue().get(key));

                        break;
                    default:
                        throw new RuntimeException("Unable to transform unknown resource of type" + key);
                }
            }
            resources.add(hwc.getInstance());
        }

        k = (int) (0.2 * model.getHardwareModel().getResourceList().asJavaCollection().size());

    }

    public Map<Integer, ArrayList<Resource>> buildClusters(){

        SimpleKMeans algorithm = new SimpleKMeans();
        Map<Integer, ArrayList<Resource>> Clusters = new HashMap<>();

        for (int i = 0; i < k; i ++)
            Clusters.put(i, new ArrayList<>());
        try {
            algorithm.setNumClusters(k);
            algorithm.setPreserveInstancesOrder(true);
            algorithm.buildClusterer(resources);

            int[] assignments = algorithm.getAssignments();

            for (int i = 0; i < assignments.length; i++){
                Clusters.get(assignments[i]).add(hardwareComponentList.get(i).getResource());
            }

        } catch (Exception e) {
            e.printStackTrace();
        }
        return Clusters;
    }
}
